/*
 *  COFF mapfile utility
 *
 *  History:
 *
 *    ?-???-?? Brian Currie
 *      Created
 *
 *    3-Aug-96 Gisle Vanem (giva@bgnett.no)
 *      Writes to map-file.
 *      Print buffers are put on stack
 *      Added NULL check on malloc()
 *      Hex values are uppercase,
 *      Added a simple mapfile header
 *      Sorted symbols by address      (-)
 *      Compiled as C file (not C++)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <dir.h>

#ifndef __GNUC__
#define __attribute__(x)
#endif

#pragma pack(1)
#include "coff.h"
#pragma pack()

typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef unsigned long  DWORD;

char   mapFile[128];
FILE   *exeF     = NULL;
FILE   *mapF     = NULL;
int    verbose   = 0;
DWORD  allocated = 0;

FILHDR coffHdr;
void   *optHdr;
SCNHDR *sections;
LINENO **lines;
SYMENT *symbols;
AUXENT *auxilary;
char   *strings;
char   *registers[] = { "eax", "ebx", "ecx", "edx",
                        "esp", "ebp", "esi", "edi"
                      };

char   *types[]     = { "????",   "void",   "char",
                        "short",  "int",    "long",
                        "float",  "double", "struct",
                        "union",  "enum",   "of enum",
                        "unsigned char",    "unsigned short",
                        "unsigned int",     "unsigned long",
                        "long double"
                      };

#ifdef __BORLANDC__   /* BCC uses 4kB stack by default */
int _stklen = 20*1024;
#endif


/*-------------------------------------------------------------------*/

static void Abort (int errLvl, char *fmt, ...)
{
  vprintf (fmt,(&fmt)+1);
  if (mapF) fclose (mapF);
  if (exeF) fclose (exeF);
  if (verbose && errLvl == 0)
     printf ("%lu bytes dynamic memory used\n",allocated);
  exit (errLvl);
}

static void * Malloc (size_t size, unsigned line)
{
  void *ret;

  if (size == 0)
     Abort (10,__FILE__" (%u): Allocating 0 bytes\n",line);

  ret = malloc(size);
  if (ret == NULL)
     Abort (10,__FILE__" (%u): Allocation of (%u bytes) failed\n",line,size);
  memset (ret,0,size);
  allocated += size;
  return ret;
}

#define malloc(size)  Malloc (size,__LINE__)


int OutF (char *fmt, ...)
{
  return vfprintf (mapF,fmt,(&fmt)+1);
}

void PrintHeader (char *fileName)
{
  OutF ("Map listing for \"%s\"\n\n"\
        "Address  Sec  Class Symbol name etc\n"
        "--------------------------------------------------------------\n",
        fileName);
}

/*-------------------------------------------------------------------*/

struct Module {
       struct Module *next;
       char          *fname;
       DWORD          fstart;
       struct {
         DWORD text;
         DWORD data;
         DWORD bss;
       } start, size;
     };

struct Module *module0 = NULL;
DWORD  numModules      = 0;

void PrintModuleInfo (void)
{
  struct Module *mod;
  DWORD  textTotal = 0L;
  DWORD  dataTotal = 0L;
  DWORD  bssTotal  = 0L;

  OutF ("\nModule                 Start     Size\n"\
          "---------------------------------------------------\n");

  for (mod = module0; mod; mod = mod->next)
  {
    OutF ("%-16s .text %08lX  %08lX (%5lu)\n"\
          "\t         .data %08lX  %08lX (%5lu)\n"\
          "\t         .bss  %08lX  %08lX (%5lu)\n",
    mod->fname, mod->start.text, mod->size.text, mod->size.text,
                mod->start.data, mod->size.data, mod->size.data,
                mod->start.bss,  mod->size.bss,  mod->size.bss);
    textTotal += mod->size.text;
    dataTotal += mod->size.data;
    bssTotal  += mod->size.bss;
  }
  OutF ("----------------------------------------------------\n"\
        "                 .text           %08lX (%6lu)\n"       \
        "                 .data           %08lX (%6lu)\n"       \
        "                 .bss            %08lX (%6lu)\n"       \
        "                                        = (%6lu)\n\n"  \
        "# of modules %d\n", textTotal,textTotal, dataTotal,dataTotal,
        bssTotal,bssTotal, textTotal + dataTotal + bssTotal, numModules);
}

int Compare (struct Module **a, struct Module **b)
{
  return ((*a)->fstart - (*b)->fstart);
}

void SortModules (unsigned num)
{
  int i;
  struct Module *mod;
  struct Module **modp = malloc (num * sizeof(**modp));

  for (i=0, mod = module0; mod; i++, mod = mod->next)
      modp[i] = mod;

  qsort (modp,num,sizeof(*mod),(int(*)())Compare);
  module0 = NULL;
  for (i=num-1; i >= 0; i--)  /* Put them back in linked list */
  {
    mod       = modp[i];
    mod->next = module0;
    module0   = mod;
  }
  free(modp);
}

/*-------------------------------------------------------------------*/

void DumpSymbol (int sym)
{
  int i,j;
  for (j = 0; j < sizeof(symbols[0]); j++)
      OutF ("%02X",*((BYTE*) &symbols[sym]+j));

  OutF ("\n");
  for (i = 1; i <= symbols[sym].e_numaux; i++)
  {
    OutF ("    ");
    for (j = 0; j < sizeof(auxilary[0]); j++)
        OutF ("%02X",*((BYTE*) &auxilary[sym+i]+j));
    OutF ("\n");
  }
}

char * SymbolName (int sym, int file)
{
  static char buf[1024];
  if (symbols[sym].e.e.e_zeroes)
       sprintf(buf,"%1.*s", file ? 14 : 8, symbols[sym].e.e_name);
  else strcpy (buf,&strings[symbols[sym].e.e.e_offset]);
  return buf;
}

char * SymbolType (int sym)
{
  static char buf[1024];
  char   tmp [2][1024];
  int    aryind = 0;
  WORD   btype  = symbols[sym].e_type;
  WORD   type   = BTYPE(btype);

  btype &= ~type;
  strcpy (tmp[0],SymbolName(sym,0));

  while (btype)
  {
    if (ISPTR(btype))
            sprintf (tmp[1],"*%s",tmp[0]);
    else if (ISFCN(btype))
            sprintf (tmp[1],"%s()",tmp[0]);
    else if (ISARY(btype))
            sprintf (tmp[1],"%s[%d]",tmp[0],
                     auxilary[sym+1].x_sym.x_fcnary.x_ary.x_dimen[aryind++]);
    else if (ISTAG(btype))
            sprintf (tmp[1],"dunno");

    strcpy (tmp[0],tmp[1]);
    btype = DECREF (btype);
  }

  switch (type)
  {
    case T_STRUCT:
    case T_UNION:
    case T_ENUM:
    case T_MOE:
         if (symbols[sym].e_numaux)
              sprintf (buf,"%s %s %s",types[type],
                       SymbolName(auxilary[sym+1].x_sym.x_tagndx,0),
                       tmp[0]);

         else sprintf (buf,"%s %s",types[type],tmp[0]);
         break;
    default:
         sprintf (buf,"%s %s",types[type],tmp[0]);
  }
  return buf;
}

int DoStruct (int sym)
{
  int end;

  OutF ("  %s %s {\n",
        symbols[sym].e_sclass == C_STRTAG ? "struct" : "union",
        SymbolName(sym,0));

  end  = auxilary[sym+1].x_sym.x_fcnary.x_fcn.x_endndx;
  sym += 1 + symbols[sym].e_numaux;
  while (sym < end)
  {
    BYTE class = symbols[sym].e_sclass;

    if (class == C_EOS)
       break;
    if (class == C_FIELD)
         OutF ("     /* bitfield */ %s;\n",
               SymbolType(sym));
    else OutF ("     /* %08lX */ %s;\n",
               symbols[sym].e_value,
               SymbolType(sym));
    sym += 1 + symbols[sym].e_numaux;
  }
  OutF ("  }; /* size=%ld bytes */\n", symbols[sym].e_value);
  sym += 1 + symbols[sym].e_numaux;
  return sym;
}

int DoEnum (int sym)
{
  int end;

  OutF ("  enum %s { /* %d ?? */\n",
        SymbolName(sym,0),
        auxilary[sym+1].x_sym.x_misc.x_lnsz.x_size);

  end  = auxilary[sym+1].x_sym.x_fcnary.x_fcn.x_endndx;
  sym += 1 + symbols[sym].e_numaux;

  while (sym < end)
  {
    if (symbols[sym].e_sclass == C_EOS)
       break;
    OutF ("    %s=%08lX,\n",
          SymbolName(sym,0),symbols[sym].e_value);
    sym += 1 + symbols[sym].e_numaux;
  }
  OutF ("  }; /* %ld ?? */\n",symbols[sym].e_value);
  sym += 1 + symbols[sym].e_numaux;

  return sym;
}

int DoTypedef (int sym)
{
  OutF ("  typedef %s;\n",SymbolType(sym));

  return (sym + 1 + symbols[sym].e_numaux);
}

int DoFile (int sym)
{
  struct Module *mod = malloc (sizeof(*mod));

  memset ((void*)mod,0,sizeof(*mod));
  mod->next   = module0;
  mod->fstart = symbols[sym].e_value;
  mod->fname  = strdup (SymbolName(sym+1,1));
  module0     = mod;
  numModules++;
  if (verbose)
     allocated += strlen(mod->fname) + 1;

  OutF ("\n.file \"%s\"\n", mod->fname);

  return (sym + 1 + symbols[sym].e_numaux);
}

int DoFunction (int sym)
{
  WORD line   = 0;
  WORD indent = 0;
  int  end;

  OutF ("%08lX %-4d       %s%s",
        symbols[sym].e_value,
        symbols[sym].e_scnum,
        symbols[sym].e_sclass == C_STAT ? "static " : "",
        SymbolType(sym));

  if (symbols[sym].e_numaux)
       OutF ("  /* size=%u */\n", auxilary[sym+1].x_sym.x_misc.x_fsize);
  else OutF ("\n");

  end  = auxilary[sym+1].x_sym.x_fcnary.x_fcn.x_endndx;
  sym += 1 + symbols[sym].e_numaux;

  while (sym < end)
  {
    if (!strcmp(symbols[sym].e.e_name,".bf"))
    {
      indent++;
      OutF ("%08lX %*s{%*s (%u)\n",
            symbols[sym].e_value,
            9 + 2 * indent, "", 30 - 2 * indent, "",
            line = auxilary[sym+1].x_sym.x_misc.x_lnsz.x_lnno);
      line--;
    }
    else if (!strcmp(symbols[sym].e.e_name,".bb"))
    {
      indent++;
      OutF ("%08lX %*s{%*s (%u)\n",
            symbols[sym].e_value,
            9 + 2 * indent, "", 30 - 2 * indent, "",
            auxilary[sym+1].x_sym.x_misc.x_lnsz.x_lnno+line);
    }
    else if (!strcmp(symbols[sym].e.e_name,".eb"))
    {
      OutF ("%08lX %*s}%*s (%u)\n",
            symbols[sym].e_value,
            9 + 2 * indent, "", 30 - 2 * indent, "",
            auxilary[sym+1].x_sym.x_misc.x_lnsz.x_lnno+line);
      indent--;
    }
    else if (!strcmp(symbols[sym].e.e_name,".ef"))
    {
      OutF ("%08lX %*s}%*s (%u)\n",
            symbols[sym].e_value,
            9 + 2 * indent, "", 30 - 2 * indent, "",
            auxilary[sym+1].x_sym.x_misc.x_lnsz.x_lnno+line);
      indent--;
    }
    else
    {
      switch (symbols[sym].e_sclass)
      {
        case C_REG  : OutF ("%*s %-*s %%%s\n",
                            19 + 2 * indent, "",
                            29 - 2 * indent,
                            SymbolType(sym),
                            registers[symbols[sym].e_value]);
                      break;
        case C_AUTO :
        case C_ARG  : OutF ("%*s %-*s (%%ebp%+ld)\n",
                            19 + 2 * indent, "",
                            29 - 2 * indent,
                            SymbolType(sym),
                            symbols[sym].e_value);
                      break;

        case C_STAT : OutF ("%*s static %s\n",
                            19 + 2 * indent, "",
                            SymbolType(sym));
                      break;

        case C_LABEL: OutF ("%*s %-*s: (%u)\n",
                            19 + 2 * indent, "",
                            29 - 2 * indent,
                            SymbolType(sym),
                            auxilary[sym+1].x_sym.x_misc.x_lnsz.x_lnno+line);
                      break;

        default     : OutF ("%*s (class %02X) %s\n",
                            19 + 2 * indent, "",
                            symbols[sym].e_sclass,
                            SymbolType(sym));
      }
    }
    sym += 1 + symbols[sym].e_numaux;
  }

  return sym;
}

int DoSymbol (int sym)
{
  int staticVariable = (symbols[sym].e_numaux == 0) &&
                       (symbols[sym].e_sclass == C_STAT);

  OutF ("%08lX %-11d %s",
        symbols[sym].e_value,
        symbols[sym].e_scnum,
        staticVariable ? "static " : "");

  if (symbols[sym].e_type)
     OutF ("%s;\n",SymbolType(sym));
  else if (symbols[sym].e_numaux)
  {
    char *sname;
    DWORD size  = auxilary[sym+1].x_scn.x_scnlen;
    DWORD start = symbols[sym].e_value;

    OutF ("%-6s /* size=%08lX numreloc=%3u lines=%u */\n",
          sname = SymbolName(sym,0), size,
          auxilary[sym+1].x_scn.x_nreloc,
          auxilary[sym+1].x_scn.x_nlinno);

    if (!strcmp(sname,_TEXT))
    {
      module0->start.text = start;
      module0->size.text  = size;
    }
    else if (!strcmp(sname,_DATA))
    {
      module0->start.data = start;
      module0->size.data  = size;
    }
    else if (!strcmp(sname,_BSS))
    {
      module0->start.bss = start;
      module0->size.bss  = size;
    }
  }
  else OutF ("%s:\n", SymbolName(sym,0));

  return (sym + 1 + symbols[sym].e_numaux);
}

void ProcessCoffSyms (void)
{
  int i;
  for (i = 0; i < coffHdr.f_nsyms; )
  {
    switch (symbols[i].e_sclass)
    {
      default:
      case C_EFCN:
      case C_NULL:
      case C_AUTO:
      case C_MOS:
      case C_ARG:
      case C_MOU:
      case C_MOE:
      case C_REGPARM:
      case C_FIELD:
      case C_AUTOARG:
      case C_LASTENT:
      case C_BLOCK:
      case C_FCN:
      case C_EOS:
           OutF ("what's this doing here? %d %d\n",
                 i,symbols[i].e_sclass);
           DumpSymbol(i);
           i += symbols[i].e_numaux;
           i++;
           break;
      case C_TPDEF:
           i = DoTypedef(i);
           break;
      case C_FILE:
           i = DoFile(i);
           break;
      case C_LINE:
      case C_ALIAS:
      case C_HIDDEN:
           DumpSymbol(i);
           i += symbols[i].e_numaux;
           i++;
           break;
      case C_STRTAG:
      case C_UNTAG:
           i = DoStruct(i);
           OutF ("\n");
           break;
      case C_ENTAG:
           i = DoEnum(i);
           OutF ("\n");
           break;
      case C_EXT:
      case C_STAT:
      case C_REG:
      case C_EXTDEF:
      case C_LABEL:
      case C_ULABEL:
      case C_USTATIC:
           if (ISFCN(symbols[i].e_type))
           {
             i = DoFunction(i);
             OutF ("\n");
           }
           else
             i = DoSymbol(i);
    }
  }

  for (i = 0; i < coffHdr.f_nscns; i++)
  {
    int j,lineNo = 0;

    OutF ("\n%1.8s\n",sections[i].s_name);
    for (j = 0; j < sections[i].s_nlnno; j++)
    {
      DWORD address;
      if (lines[i][j].l_lnno)
      {
        address = lines[i][j].l_addr.l_paddr;
        OutF ("%08lX %-5d\n",
              address,
              lines[i][j].l_lnno+lineNo);
      }
      else
      {
        int sym = lines[i][j].l_addr.l_symndx;
        address = symbols[sym].e_value;
        lineNo  = auxilary [sym+1+symbols[sym].e_numaux+1].x_sym.x_misc.x_lnsz.x_lnno - 1;
        OutF ("%08lX %-5d %s\n",
              address,lines[i][j].l_lnno+lineNo,
              SymbolType(lines[i][j].l_addr.l_symndx));
      }
    }
  }
}

void ProcessCoff (char *file, FILE *fd, long ofs)
{
  int    i;
  DWORD strsize;

  fseek (fd,ofs,0);
  fread (&coffHdr,1,FILHSZ,fd);

  if (coffHdr.f_opthdr)
  {
    optHdr = malloc(coffHdr.f_opthdr);
    fread (optHdr,1,coffHdr.f_opthdr,fd);
  }

  sections = (SCNHDR*) malloc(coffHdr.f_nscns*SCNHSZ);
  fread (sections,coffHdr.f_nscns,SCNHSZ,fd);

  lines = (LINENO**) malloc(coffHdr.f_nscns*sizeof(LINENO*));

  for (i = 0; i < coffHdr.f_nscns; i++)
  {
    if (sections[i].s_nlnno)
    {
      lines[i] = (LINENO*) malloc(sections[i].s_nlnno*LINESZ);
      fseek (fd,ofs+sections[i].s_lnnoptr,0);
      fread (lines[i],sections[i].s_nlnno,LINESZ,fd);
    }
  }

  fseek (fd,ofs+coffHdr.f_symptr+coffHdr.f_nsyms*SYMESZ,0);
  fread (&strsize,1,4,fd);
  strings = (char *) malloc(strsize);
  fread (strings+4,1,strsize-4,fd);
  memcpy (strings,&strsize,4);

  fseek (fd,ofs+coffHdr.f_symptr,0);
  if (coffHdr.f_nsyms == 0)
     Abort (0,"File has no symbol table");

  symbols = (SYMENT*) malloc(coffHdr.f_nsyms*SYMESZ);
  fread (symbols,SYMESZ,coffHdr.f_nsyms,fd);
  auxilary = (AUXENT*)symbols;

  mapF = fopen (mapFile,"wt");
  if (!mapF)
     Abort (3,"Cannot open %s\n",mapFile);

  PrintHeader (file);
  ProcessCoffSyms();
//SortModules (numModules);  // not finished
  PrintModuleInfo();
}

void ProcessFile (char *file, FILE *fd, long ofs)
{
  short s, exe[2];

  fseek (fd,ofs,0);
  fread (&s,1,2,fd);

  switch (s)
  {
    case 0x5A4D:    /* .exe */
         fread (exe,2,2,fd);
         ofs += (long)exe[1] * 512L;
         if (exe[0])
            ofs += (long)exe[0] - 512L;
         ProcessFile (file,fd, ofs);
         break;
    case 0x014C:    /* .coff */
         ProcessCoff (file,fd,ofs);
         break;
    case 0x010B:    /* a.out ZMAGIC */
    case 0x0107:    /* a.out object */
        /* ProcessAout (file,fd, ofs); */
         Abort (4,"A.OUT unsupported\n");
         break;
    default:
         Abort (5,"Unknown file type %s\n",file);
  }
}

int main (int argc, char **argv)
{
  int i = 1;

  if (argc < 2)
     Abort (1,"Usage: %s [-v] exe-file [map-file]\n",argv[0]);

  if (argv[1][0] == '-' || argv[1][0] == '/')
  {
    if (argv[1][1] == 'v')
         verbose = 1, i++;
    else Abort (1,"Illegal option\nUsage: %s [-v] exe-file [map-file]\n",
                argv[0]);
  }

  exeF = fopen (argv[i],"rb");
  if (!exeF)
     Abort (2,"Cannot open %s\n",argv[i]);

  if (argc == i+2)
     strcpy (mapFile,argv[i+1]);
  else
  {
    char tmp[MAXDIR];
    fnsplit (argv[i],tmp,tmp,mapFile,tmp);
    strcat (mapFile,".map");
  }

  if (verbose)
     puts ("GNU-map v0.2");

  ProcessFile (argv[i],exeF,0);
  Abort (0,"");
  return 0;
}
